#################################################################################
# NOTICE                                                                        #
# This software script is protected by                                          #
# the Copyright Act of 1976, 17 U.S.C. $$ 101-810, as amended. Rights reserved. #
# Please contact The Brigham and Women's Hospital, Inc. for further information.#
#################################################################################
# The deveroper and the maintainer: Lang Ho Lee (lhlee@bwh.harvard.edu)         #
#                                   Sasha A. Singh (sasingh@bwh.harvard.edu)    #
#################################################################################

# Mute warnings
options(warn=1)

#' @title make_random_xina_data
#' @description Generate random proteomics dataset for testing XINA
#' 'make_random_xina_data' will make random proteomics data for XINA test.
#' The generated data will have three conditions and seven time points,
#' c("0hr", "2hr", "6hr", "12hr", "24hr", "48hr", "72hr").
#' @param n The number of proteins for one condition. Default is 500.
#' @param mtor If it is TRUE (default), mTOR pathway genes will be significant.
#' If it is FALSE, randomly selected genes will be significant in first three conditions.
#' @param time_points A vector containing time points of the data matrix
#' @param conditions A vector containing condition information, for example normal, disease and drug treated disase.
#' @return Three comma-separated files containing time-series data for XINA
#' @export
#' @examples
#' \dontrun{
#' library(XINA)
#' make_random_xina_data()
#' g1 <- read.csv("Control.csv", check.names=FALSE, stringsAsFactors = FALSE)
#' head(g1)
#' g2 <- read.csv("Stimulus1.csv", check.names=FALSE, stringsAsFactors = FALSE)
#' head(g2)
#' g3 <- read.csv("Stimulus2.csv", check.names=FALSE, stringsAsFactors = FALSE)
#' head(g3)
#'
#' # Model-based clustering
#' example_clusters <- xina_clustering(c("Control.csv","Stimulus1.csv","Stimulus2.csv"),
#' data_column=c("0hr","6hr","12hr","24hr","48hr","72hr"), nClusters=30)
#' head(example_clusters$aligned)
#'
#' # Plot clustered plots
#' plot_clusters(example_clusters)
#'
#' # Plot pie-charts of condition composition
#' condition_composition <- plot_condition_compositions(example_clusters)
#' }
#'
make_random_xina_data <- function(n=500, mtor=TRUE,
                                  time_points=c("0hr", "2hr", "6hr", "12hr", "24hr", "48hr", "72hr"),
                                  conditions=c("Control", "Stimulus1", "Stimulus2")) {
  # Get gene names and descriptions
  data(gene_info)
  # Get mTOR pathway data
  if (mtor) {
    list_intended_genes <- get_mTOR_proteins(time_points, conditions)
    nrow_intended <- nrow(list_intended_genes[[1]])
  } else {
    nrow_intended <- 0
  }
  # Calculate required numbers
  num_total <- n - nrow_intended
  num_shared <- round(num_total*0.8)
  num_unique <- num_total - num_shared
  limit_gn <- length(gn) - nrow_intended
  # Get random data
  if (mtor) {
    list_data <- get_random_data(time_points, conditions, num_shared+num_unique, percent.sign=0)
  } else {
    list_data <- get_random_data(time_points, conditions, num_shared+num_unique)
  }
  # Get random gene names
  shared_idx <- sample(c(1:limit_gn), num_shared)
  for (i in 1:length(conditions)) {
    selected_idx <- shared_idx
    for (j in 1:num_unique) {
      tmp_idx <- 0
      while (TRUE) {
        tmp_idx <- sample(c(1:limit_gn), 1)
        if (is.na(match(tmp_idx, selected_idx))) {
          break
        }
      }
      selected_idx <- c(selected_idx, tmp_idx)
    }
    if (mtor) {
      data_out <- rbind(cbind(Accession=gn[selected_idx], Description=desc[selected_idx], list_data[[i]]),
                        list_intended_genes[[i]])
    } else {
      data_out <- cbind(Accession=gn[selected_idx], Description=desc[selected_idx], list_data[[i]])
    }
    write.csv(data_out, paste(conditions[i],".csv",sep=''), row.names = FALSE)
  }
  return(list(size=n, time_points=time_points, conditions=conditions))
}

#' @title get_random_data
#' @description Get randomized time-series data
#' @param time_points A vector containing time points of the data matrix
#' @param conditions A vector containing condition information, for example normal, disease and drug treated disase.
#' @param num_total The number of total proteins to be generated
#' @param percent.sign Percentage of differentially expressed proteins. Ignored when equal=FALSE.
#' @param equal If equal is TRUE, all the conditions will have numbers between 0 and 1.
#' If it is  FALSE, the first three conditions will have different ranges.
#' First condition will have numbers from 0.3 to 0.4.
#' Second condition will have numbers from 0.6 to 0.8.
#' Third condition will have numbers from 0.3 to 0.5.
#' Other conditions will have numbers from 0 to 1.
#' @return A list containing ramdomly generated data matrix
#'
get_random_data <- function(time_points, conditions, num_total, percent.sign=0.1, equal=TRUE){
  if (equal) {
    significant_idx <- sample(1:num_total, round(num_total*percent.sign))
  } else {
    significant_idx <- c(1:num_total)
  }
  list_data <- list()
  for (i in 1:num_total) {
    for (j in 1:length(conditions)) {
      if (is.na(match(2,significant_idx))) {
        expression_profile <- c(0.5, runif(length(time_points)-1))
      } else {
        if (j==1) {
          expression_profile <- c(0.5, runif(length(time_points)-1,0.3,0.4))
        } else if (j==2) {
          expression_profile <- c(0.5, runif(length(time_points)-1,0.6,0.8))
        } else if (j==3) {
          expression_profile <- c(0.5, runif(length(time_points)-1,0.3,0.5))
        } else {
          expression_profile <- c(0.5, runif(length(time_points)-1))
        }
      }
      names(expression_profile) <- time_points
      if (length(list_data)<j) {
        list_data[[j]] <- round(expression_profile,4)
      } else {
        tmp <- rbind(list_data[[j]], round(expression_profile,4))
        colnames(tmp) <- time_points
        rownames(tmp) <- NULL
        list_data[[j]] <- tmp
      }
    }
  }
  return(list_data)
}

#' @title get_mTOR_proteins
#' @description Get mTOR pathway genes
#' @param time_points A vector containing time points of the data matrix
#' @param conditions A vector containing condition information, for example normal, disease and drug treated disase.
#' @return A vector containing mTOR pathway gene names
get_mTOR_proteins <- function(time_points, conditions) {
  data(gene_info)
  mtor_gn <- c("AKT1","AKT1S1","BRAF","CAB39","CAB39L","CBX2",
               "DDIT4","EIF4B","EIF4E","EIF4E2","EIF4EBP1",
               "HIF1A","IGF1","IKBKB","IRS1","MAPK1","MAPK3","MLST8",
               "MTOR","PDPK1","PIK3CA","PIK3CB","PIK3CD","PIK3CG",
               "PIK3R1","PIK3R2","PIK3R3","PIK3R5","PRKAA1","PRKAA2",
               "PRKCA","PRKCB","PRKCG","PTEN","RHEB","RICTOR","RPS6",
               "RPS6KA1","RPS6KA2","RPS6KA3","RPS6KA6","RPS6KB1",
               "RPS6KB2", "RPTOR","RRAGA","RRAGB","RRAGC","RRAGD","RYBP",
               "STK11","STRADA", "TNF","TSC1","TSC2","ULK1","ULK2","VEGFA")
  mtor_desc <- desc[match(mtor_gn, gn)]
  list_data <- get_random_data(time_points, conditions, length(mtor_gn), equal=FALSE)
  list_return <- list()
  for (i in 1:length(conditions)) {
    list_return[[i]] <- cbind(Accession=mtor_gn, Description=mtor_desc, list_data[[i]])
  }
  return(list_return)
}

#' @title extract_data_column
#' @description Extract data column names from XINA clustering result
#' @param col_head_of_clustering Column names of XINA clustering result
#' @return A vector containing column names of  data matrix
extract_data_column <- function(col_head_of_clustering){
  cluster_info_head <-c("Accession", "Description", "Condition", "Key", "Cluster")
  data_column <- c()
  for (col_head in col_head_of_clustering){
    if (is.na(match(col_head, cluster_info_head))){
      data_column <- c(data_column, col_head)
    }
  }
  return(data_column)
}

#' @title load_previous_results
#' @description Get previous XINA clustering results to R space
#' @param clustering_dir The directory path of XINA clustering results
#' @param data_column A vector containing column names of  data matrix
#' @param fp_clusters File path of XINA clustering results
#' @return Comma-separated file containing aligned XINA clustering results.
#' @export
#' @examples
#' \dontrun{
#' library(XINA)
#'
#' # Generate random multiplexed time-series data
#' random_data_info <- make_random_xina_data()
#' # 'nClusters' is the number of desired clusters.
#' # 'mclust' will chose the most optimized number of clusters
#' # considering Bayesian information criteria (BIC). BIC of 'mclust' is negative BIC,
#' # thus the higher BIC, the more optimized clustering.
#' # Data files
#' data_files <- paste(random_data_info$conditions, ".csv", sep='')
#'
#' # time points of the data matrix
#' data_column <- random_data_info$time_points
#'
#' # mclust requires the fixed random seed to get reproduce the clustering results
#' set.seed(0)
#'
#' # Run the model-based clustering
#' example_clusters <- xina_clustering(data_files, data_column=data_column, nClusters=30)
#'
#' # Reload the clustering result
#' example_clusters_reloaded <- load_previous_results(".")
#' head(example_clusters_reloaded$aligned)
#' }
#'
load_previous_results <- function(clustering_dir=getwd(), data_column=NULL,
                                  fp_clusters="xina_clusters.csv") {
  # Retrieve clustering results
  file_clusters <- paste(clustering_dir, "/", fp_clusters, sep='')
  df_clusters <- read.csv(file_clusters, check.names=FALSE, stringsAsFactors = FALSE)
  df_aligned <- organize_clusters(clustering_dir, df_clusters, file_out=FALSE)
  if (is.null(data_column)) {
    data_column <- extract_data_column(colnames(df_clusters))
  }
  conditions <- unique(df_clusters$Condition)
  condition_colors <- get_color_for_nodes()[1:length(conditions)]
  names(condition_colors) <- conditions
  nClusters <- max(df_clusters$Cluster)
  color_for_clusters <- get_colors(nClusters, set='original')
  names(color_for_clusters) <- c(1:nClusters)
  # Recall previous results
  mIMT_result <- list()
  mIMT_result$clusters <- df_clusters
  mIMT_result$aligned <- df_aligned
  rownames(mIMT_result$clusters) <- mIMT_result$clusters$Key
  mIMT_result$data_column <- data_column
  mIMT_result$out_dir <- clustering_dir
  mIMT_result$nClusters <- nClusters
  mIMT_result$max_cluster <- nClusters
  mIMT_result$chosen_model <- NA
  mIMT_result$optimal_BIC <- NA
  mIMT_result$condition <- conditions
  mIMT_result$color_for_condition <- condition_colors
  mIMT_result$color_for_clusters <- color_for_clusters
  return(mIMT_result)
}


#' @title organize_clusters
#' @description Compare clusters and find similar ones
#' @param clustering_dir The directory path of XINA clustering results
#' @param super_ds XINA clusters
#' @param file_out If it is TRUE, it writes the aligned clustering informaion to "xina_clusters_aligned.csv" file.
#' @return Comma-separated file containing aligned XINA clustering results.
organize_clusters <- function(clustering_dir=getwd(), super_ds, file_out=TRUE)
{
  conditions <- unique(super_ds$Condition)
  genes <- unique(super_ds$Accession)

  df_cluster <- data.frame()
  for (i in 1:length(genes))
  {
    ds_gene <- subset(super_ds, Accession == genes[i])
    clusters <- c()
    for (j in 1:length(conditions))
    {
      cluster <- subset(ds_gene, Condition==conditions[j])
      if (nrow(cluster)>0)
      {
        cluster_number <- as.integer(cluster$Cluster)
      } else {
        cluster_number <- 0
      }
      clusters <- c(clusters, cluster_number)
    }
    tmp <- c(as.character(ds_gene$Accession[1]), as.character(ds_gene$Description[1]), clusters)
    df_cluster <- rbind(df_cluster, data.frame(t(tmp), stringsAsFactors = FALSE))
  }
  colnames(df_cluster) <- c("Gene name", "Description", as.vector(conditions))
  if (file_out) {
    write.csv(df_cluster,"xina_clusters_aligned.csv",row.names = FALSE)
  }
  return(df_cluster)
}

#' @title xina_clustering
#' @description Clustering multiplexed time-series omics data to find co-abundance profiles
#' @param f_names A vector containing input file (.csv) paths
#' @param data_column A vector containing column names (1st row of the input file) of data matrix
#' @param out_dir A directory path for saving clustering results. (default: out_dir=getwd())
#' @param nClusters The number of desired maximum clusters
#' @param chosen_model \link[mclust]{mclustModelNames}
#' @param norm Default is "sum_normalization".  Sum-normalization is to divide the data matrix by row sum.
#' If you want to know more about sum-normalization, see https://www.ncbi.nlm.nih.gov/pubmed/19861354.
#' "zscore" is to calculate Z score for each protein. See \link[base]{scale}.
#' @return a plot containing a BIC plot in current working directory and a list containing below information:
#'      \tabular{rl}{
#'       \strong{Item} \tab \strong{Description}\cr
#'       clusters \tab XINA clustering results\cr
#'       aligned \tab XINA clustering results aligned by ID\cr
#'       data_column \tab Data matrix column names\cr
#'       out_dir \tab The directory path containing XINA results\cr
#'       nClusters \tab The number of clusters desired by user\cr
#'       max_cluster \tab The number of clusters optimized by BIC\cr
#'       chosen_model \tab The used covariance model for model-based clustering\cr
#'       optimal_BIC \tab BIC of the optimized covariance model\cr
#'       condition \tab Experimental conditions of the user input data\cr
#'       color_for_condition \tab Colors assigned to each experimental conditions which is used for condition composition plot\cr
#'       color_for_clusters \tab Colors assigned to each clusters which is used for XINA clustering plot\cr
#'       norm_method \tab Used normalization method\cr
#'      }
#' @export
#' @examples
#' \dontrun{
#' library(XINA)
#'
#' # Generate random multiplexed time-series data
#' random_data_info <- make_random_xina_data()
#' # 'nClusters' is the number of desired clusters.
#' # 'mclust' will chose the most optimized number of clusters
#' # considering Bayesian information criteria (BIC). BIC of 'mclust' is negative BIC,
#' # thus the higher BIC, the more optimized clustering.
#' # Data files
#' data_files <- paste(random_data_info$conditions, ".csv", sep='')
#'
#' # time points of the data matrix
#' data_column <- random_data_info$time_points
#'
#' # mclust requires the fixed random seed to get reproduce the clustering results
#' set.seed(0)
#'
#' # Run the model-based clustering to find co-abundance profiles
#' example_clusters <- xina_clustering(data_files, data_column=data_column, nClusters=30)
#' }
#'
xina_clustering <- function(f_names, data_column, out_dir=getwd(), nClusters=20,
                            norm="sum_normalization", chosen_model="")
{
  clustering_dir <- getwd()
  # Merge data
  superset <- generate_superset(f_names, data_column, norm=norm)
  rownames(superset) <- superset$Key
  # Run Mclust
  if (chosen_model == "") {
    set.seed(0)
    mc_Sds <- Mclust(superset[data_column],G=2:nClusters)
    chosen_model <- mc_Sds$modelName
  } else {
    mc_Sds <- Mclust(superset[data_column],G=2:nClusters, modelNames=chosen_model)
  }
  superset$Cluster <- mc_Sds$class
  plot.Mclust(mc_Sds, what="BIC")

  # Write messages from Mclust
  sink(paste("./mclust_summary.txt", sep=''))
  summSds<-summary(mc_Sds,parameters=T)
  print(summSds)
  sink()

  # Remove empty clusters
  superset_nr <- data.frame()
  cluster_num <- 0
  for (i in 1:mc_Sds$G){
    sub_cluster <- subset(superset, Cluster==i)
    if (nrow(sub_cluster) > 0){
      cluster_num <- cluster_num + 1
      sub_cluster[,"Cluster"] <- cluster_num
      superset_nr <- rbind(superset_nr, sub_cluster)
    }
  }
  write.csv(superset_nr, file="xina_clusters.csv", row.names = FALSE)
  # Collect cluster information by proteins
  aligned_clusters <- organize_clusters(clustering_dir, superset_nr)
  conditions <- unique(superset_nr$Condition)
  max_cl <- max(superset_nr$Cluster)
  # Set default condition colors
  condition_colors <- get_color_for_nodes()[1:length(conditions)]
  names(condition_colors) <- conditions
  # Set default cluster colors
  color_for_clusters <- get_colors(nClusters, set='original')
  names(color_for_clusters) <- c(1:nClusters)
  # return the return list
  return(list("clusters"=superset_nr,
              "aligned"=aligned_clusters,
              "data_column"=data_column,
              "out_dir"=clustering_dir,
              "nClusters"=nClusters,
              "max_cluster"=max(superset_nr$Cluster),
              "chosen_model"=chosen_model,
              "optimal_BIC"=mc_Sds$bic,
              "condition"=conditions,
              "color_for_condition"=condition_colors,
              "color_for_clusters"=color_for_clusters,
              "norm_method"=norm))
}

#' @title generate_superset
#' @description Merge input kinetics files
#' @param f_names A vector of .csv file paths containing kinetics data
#' @param data_column A vector of column names containing data matrix
#' @param delim The delimiter of input file (default is ',')
#' @param norm The normalization method. It should be one of c('sum_normalization', 'zscore').
#' Default is 'sum_normalization'.
#' @export
#' @return A data frame containing kinetics data obtained from files in the f_names vector
generate_superset <- function(f_names, data_column, delim=",", norm="sum_normalization")
{
  SuperDS <- data.frame()
  for (i in 1:length(f_names))
  {
    f_name <- f_names[i]
    condition <- file_path_sans_ext(f_name)
    df_csv <- read.csv(file=f_name, check.names=FALSE, sep=delim)
    if (is.na(match("Accession",colnames(df_csv)))){
      stop(paste("Your input file named '",f_name,"' doesn't have 'Accession' column"))
    }
    if (is.na(match("Description",colnames(df_csv)))){
      stop(paste("Your input file named '",f_name,"' doesn't have 'Description' column"))
    }
    # remove zero-sum rows
    df_filtered <- subset(df_csv, rowSums(df_csv[data_column])>0)
    # remove na rows
    df_filtered <- df_filtered[complete.cases(df_filtered[data_column]), ]
    # normalize the data matrix
    if (norm=="sum_normalization") {
      # Sum-normalization
      df_normalized <- cbind(df_filtered[data_column] / as.matrix(rowSums(df_filtered[data_column])),
                             Accession=df_filtered$Accession, Description=df_filtered$Description, Condition=condition)
    } else if (norm=="zscore") {
      # Z-score
      df_normalized <- data.frame()
      for (k in 1:nrow(df_filtered)) {
        tmp <- as.vector(scale(as.numeric(df_filtered[k, data_column])))
        names(tmp) <- data_column
        df_normalized <- rbind(df_normalized, tmp)
        colnames(df_normalized) <- data_column
      }
      df_normalized["Accession"] <- df_filtered$Accession
      df_normalized["Description"] <- df_filtered$Description
      df_normalized["Condition"] <- condition
    } else {
      stop("normalization method should be one of these c(sum_normalization, zscore)")
    }
    # Add unique keys
    df_normalized["Key"] <- paste(df_normalized$Accession, "@", condition, sep="")
    # Remove duplicated keys
    df_normalized=df_normalized[!duplicated(df_normalized$Key),]
    # append the normalized row
    SuperDS <-rbind(SuperDS, df_normalized)
  }
  return(SuperDS)
}
